// bc_smg.cpp : Defines the entry point for the DLL application.
//
// Generates a surface map for a specified X and Y coordinate.
// The surface map generated is returned as "mapdata".
// There is by no means anything spectacular about this
// generator.
//
// BattleCraft 1942 and BattleField 1942 are registered trademarks of
// Digital Illusions CE.  (c)2003 by Digital Illusions CE.
// All Right Reserved.
//
// Programmed by Eric PL Smith (aka Zerk) of Red Monkey Studios
//
/* MATERIAL MAP DATA BY TYPE
0 Default
1 Water/Ice/Slippery
2 Dry Grass
3 Wet Grass
4 Dry Dirt
5 Wet Dirt
6 Mud
7 Outside Map
8 Gravel
9 Frozen Ground
10 Dry Sand
11 Wet Sand
12 Rock Surface
13 Sand Road
14 Dirt Road
15 Paved Road
*/

#include "stdafx.h"
#include "resource.h"
#include <malloc.h>
#include <stdlib.h>

#include "bc_smg.h"

LRESULT CALLBACK DialogWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);

HINSTANCE hInstance;

surfacebmps_s surfacebmps[16];

char module[MAXMODULE];

int detail = 2; //the detail level to generate.  You can set this value with a dialog if you desire.

matmap_s mmap[1024+9][1024+9];
int MAX = 0;
int texOffsetX = 0;
int texOffsetY = 0;

int smx;
int smy;

//The main entry point of the DLL.  This is called when the DLL is
//loaded and is called when BattleCraft 1942 first starts up.
BOOL APIENTRY DllMain( HINSTANCE hinst, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	hInstance = hinst;
    return TRUE;
}

//initSurfaceMap()
// This is called then the user selects the "Generate Surface Maps" option
// from within BC42 and is called by BC42.  Return TRUE to continue the
// generation process or return FALSE and BC42 will abort the process.
//
// matmap = a pointer to the material map data stored in a multidimensional array
// max = the max array size of the map (256, 512, 1024)
// offx, offy = Map offsets
//
extern "C" __declspec(dllexport) bool initSurfaceMap(matmap_s *matmap, int max, int offx, int offy) {
	bool result = true;

	//user defined dialog.
	int ret = DialogBox(hInstance, (LPCTSTR)IDD_SMG_DIALOG, NULL, (DLGPROC)DialogWndProc);
	if(ret != IDOK)
		return false;

	//set variables used in the generation
	MAX = max;
	texOffsetX = offx;
	texOffsetY = offy;

	//copy the material map data to a new memory location
	memcpy(mmap, matmap, (1033*1033)*sizeof(matmap_s));

	return result;
}

// initSurfaceBmp()
// Called by BC42 to store the loaded bitmap information that is
// the TextureSet set by the user so that this DLL can use the
// correct bitmaps to use in the generation.
//
// index = the number of the surfacebmp array to store
// surfbmp = the surface bmp data as stored by BC42
//
extern "C" __declspec(dllexport) void initSurfaceBmp(int index, surfacebmps_s *surfbmp) {
	memcpy(&surfacebmps[index], surfbmp, sizeof(surfacebmps_s));
}

// generateSurfaceMap()
// This is called for each surface map grid selected by the user via
// the Surface Map Assignments dialog in BC42
//
// size = The size of the map to generate.  BC42 passes 1024.
// gridx = The X map grid coordinate.
// gridy = The Y map grid coordinate.
//
extern "C" __declspec(dllexport) unsigned char *generateSurfaceMap(int size, int gridx, int gridy) {
	//allocate the memory needed to store our surface map
	//BC42 will turn this into a .DDS file and store it in the level
	//RFA upon return. (1024x1024 24-bit image file)
	unsigned char *mapdata = (unsigned char*)malloc(size*size*3);

	//get the name of the DLL module called
	GetModuleFileName(NULL, (LPTSTR)module, MAXMODULE);

	//generate the surface map...
	generateLowDetail(mapdata, gridx, gridy);
	if(detail > 0) generateMediumDetail(mapdata, gridx, gridy);
	if(detail > 1) generateHighDetail(mapdata, gridx, gridy);

	//return our surface map back to BC42
	return mapdata;
}

void generateLowDetail( unsigned char *map, int dx, int dy ) {
	int ds = 0;
	if(MAX == 1024)
		ds = 16;
	else if(MAX == 512)
		ds = 8;
	else if (MAX == 256)
		ds = 4;

	int smmx = (MAX/ds)*texOffsetX;
	int smmy = (MAX/ds)*texOffsetY;
	smmx += (dx*(MAX/ds));
	smmy += (dy*(MAX/ds));
	int smwh = MAX/ds;

	int pixsize = 16;

	int mdx = 0;
	int mdy = 0;
	int smd = 0;
	int smdo = 0;
	int v = 0;
	int va = 0;
	int v1 = 0;
	int v2 = 0;
	int v3 = 0;
	int mx = smmx;
	int my = smmy;

	for (my = smmy; my < smmy+smwh; my++)
	{
		for (mx = smmx; mx < smmx+smwh; mx++)
		{
			int offx = mx-smmx;
			int offy = my-smmy;

			va = mmap[MAX-mx][my].type;
			if (my > 0)
				v1 = mmap[MAX-mx][my-1].type;
			if (mx > 0)
				v2 = mmap[(MAX-mx)+1][my].type;
			if (my <= MAX)
				v3 = mmap[MAX-mx][my+1].type;
			
			smdo = surfacebmps[v].w/pixsize;

			bool join = false;
			if(v1 == v2)
				join = true;
			bool join2 = false;
			if(v2 == v3)
				join2 = true;

			int sect = pixsize;
			for (int by = 0; by < pixsize; by++)
			{
				for (int bx = 0; bx < pixsize; bx++)
				{
					if(join && bx < sect)
					{
						v = v1;
					}
					else
						v = va;

					smd = (bx*3)+(by*(surfacebmps[v].w*3)) + ((smx*(pixsize*3))+(smy*((surfacebmps[v].h*3)*pixsize)));
					mdx = (bx*3)+(by*(1024*3)) + ((offx*(pixsize*3))+(offy*((1024*3)*pixsize)));

					map[mdx] = surfacebmps[v].data[smd];
					map[mdx+1] = surfacebmps[v].data[smd+1];
					map[mdx+2] = surfacebmps[v].data[smd+2];
				}
				sect--;
			}

			if (join2)
			{
				sect = 0;
				for (int by = 0; by < pixsize; by++)
				{
					for (int bx = 0; bx < pixsize; bx++)
					{
						if(bx <= sect)
							v = v3;
						else
							v = va;

						smd = (bx*3)+(by*(surfacebmps[v].w*3)) + ((smx*(pixsize*3))+(smy*((surfacebmps[v].h*3)*pixsize)));
						mdx = (bx*3)+(by*(1024*3)) + ((offx*(pixsize*3))+(offy*((1024*3)*pixsize)));

						map[mdx] = surfacebmps[v].data[smd];
						map[mdx+1] = surfacebmps[v].data[smd+1];
						map[mdx+2] = surfacebmps[v].data[smd+2];
					}
					sect++;
				}
			}

			smx++;
			if (smx >= smdo)
			{
				smx = 0;
			}
		}
		smy++;
		if (smy >= smdo)
		{
			smy = 0;
		}
	}
}

void generateMediumDetail( unsigned char *map, int dx, int dy  ) {
	int ds = 0;
	if(MAX == 1024)
		ds = 16;
	else if(MAX == 512)
		ds = 8;
	else if (MAX == 256)
		ds = 4;

	int smmx = (MAX/ds)*texOffsetX;
	int smmy = (MAX/ds)*texOffsetY;
	smmx += (dx*(MAX/ds));
	smmy += (dy*(MAX/ds));
	int smwh = MAX/ds;

	int mdx = 0;
	int mdy = 0;
	int smd = 0;
	int smdo = 0;
	int v = 0;
	int va = 0;
	int v1 = 0;
	int v2 = 0;
	int v3 = 0;
	int mx = smmx;
	int my = smmy;

	for (my = smmy; my < smmy+smwh; my++)
	{
		for (mx = smmx; mx < smmx+smwh; mx++)
		{
			int offx = mx-smmx;
			int offy = my-smmy;

			va = mmap[MAX-mx][my].type;
			if (mx > 0)
				v1 = mmap[(MAX-mx)+1][my].type;
			if (my > 0 && mx > 0)
				v2 = mmap[(MAX-mx)+1][my-1].type;
			if (my > 0)
				v3 = mmap[MAX-mx][my-1].type;
			
			smdo = surfacebmps[v].w/16;

			for (int t = 0; t < 50; t++)
			{
				//above
				int bx = (int)(rand()%16);
				int by = (int)(rand()%16);

				v = v3;

				smd = (bx*3)+(by*(surfacebmps[v].w*3)) + ((smx*(16*3))+(smy*((surfacebmps[v].h*3)*16)));
				mdx = (bx*3)+(by*(1024*3)) + ((offx*(16*3))+(offy*((1024*3)*16)));

				map[mdx] = surfacebmps[v].data[smd];
				map[mdx+1] = surfacebmps[v].data[smd+1];
				map[mdx+2] = surfacebmps[v].data[smd+2];

				//kitty corner
				bx = (int)(rand()%8);
				by = (int)(rand()%8);

				v = v2;

				smd = (bx*3)+(by*(surfacebmps[v].w*3)) + ((smx*(16*3))+(smy*((surfacebmps[v].h*3)*16)));
				mdx = (bx*3)+(by*(1024*3)) + ((offx*(16*3))+(offy*((1024*3)*16)));

				map[mdx] = surfacebmps[v].data[smd];
				map[mdx+1] = surfacebmps[v].data[smd+1];
				map[mdx+2] = surfacebmps[v].data[smd+2];


				//left
				bx = (int)(rand()%16);
				by = (int)(rand()%16);

				v = v1;

				smd = (bx*3)+(by*(surfacebmps[v].w*3)) + ((smx*(16*3))+(smy*((surfacebmps[v].h*3)*16)));
				mdx = (bx*3)+(by*(1024*3)) + ((offx*(16*3))+(offy*((1024*3)*16)));

				map[mdx] = surfacebmps[v].data[smd];
				map[mdx+1] = surfacebmps[v].data[smd+1];
				map[mdx+2] = surfacebmps[v].data[smd+2];
			}

			smx++;
			if (smx >= smdo)
			{
				smx = 0;
			}
		}
		smy++;
		if (smy >= smdo)
		{
			smy = 0;
		}
	}
}

void generateHighDetail( unsigned char *map, int dx, int dy  ) {
}



LRESULT CALLBACK DialogWndProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_INITDIALOG:
			{
				return TRUE;
			}
		case WM_DESTROY:
			{
				EndDialog(hDlg, IDOK);
				return TRUE;
			}

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
			case IDOK:
				{
					EndDialog(hDlg, IDOK);
					return TRUE;
				}
			case IDCANCEL:
				{
					EndDialog(hDlg, IDCANCEL);
					return TRUE;
				}
			}
			break;
	}
    return FALSE;
}

